{$S-,R-,V-,I-,B-,F+,O+,A-}
Unit Timer;
  { Unit for BIOS timing functions }

INTERFACE

Type
  { For calculating timeouts }
  EventTimer = record
    StartTics : LongInt;
    ExpireTics : LongInt;
  end;

Const
  TicsPerDay = 1573040; { Assumes 18.20648 tics/sec    }
  SecsPerDay = 86400;   { Number of seconds in one day }

  { Clock frequency of 1193180/65536 is reduced to 1675/92. This }
  { allows longint conversions of tics values upto TicsPerDay    }
  TicsFreq = 1675;
  SecsFreq = 92;

  { Fast machines, >= 486-33, require an extra loop for delays }
  ExtraLoop = 25;

var
  CountsPerMs : Word; { Tight loop count for one millisecond }

Function Tics2Secs(Tics : LongInt) : LongInt;
  { Returns seconds value for Tics tics }

Function Secs2Tics(Secs : LongInt) : LongInt;
  { Returns tics value for Secs seconds }

Procedure NewTimer(var ET : EventTimer; Tics : LongInt);
  { Returns a set EventTimer that will expire in Tics }

Procedure NewTimerSecs(var ET : EventTimer; Secs : LongInt);
  { Returns a set EventTimer that will expire in Secs seconds }

Function TimerExpired(ET : EventTimer) : Boolean;
  { Returns True if ET has expired }

Function ElapsedTime(ET : EventTimer) : LongInt;
  { Returns elapsed time, in tics, for this timer }

Function ElapsedTimeInSecs(ET : EventTimer) : LongInt;
  { Returns elapsed time, in seconds, for this timer }

Function ElapsedTimeInMSecs(ET : EventTimer) : LongInt;
  { Returns elapsed time, in milliseconds, for this timer }

Function RemainingTime(ET : EventTimer) : LongInt;
  { Returns remaining time, in tics, for this timer }

Function RemainingTimeInSecs(ET : EventTimer) : LongInt;
  { Returns remaining time, in seconds, for this timer }

Function RemainingTimeInMSecs(ET : EventTimer) : LongInt;
  { Returns remaining time, in milliseconds, for this timer }

Procedure DelayTics(Tics : LongInt);
  { Delay for Tics tics }

Procedure Delay(Ms : Word);
  { Delay for specified number of milliseconds (same as CRT unit Delay) }

Procedure CalibrateDelay;
  { Delay calibration routine, similar to CRT unit }

IMPLEMENTATION

Var
  { Data area for BIOS tics value }
  BiosTics : ^LongInt;
  BiosTicsLow : ^Word;

Function Tics2Secs(Tics : LongInt) : LongInt;
  { Returns seconds value for Tics tics }
begin
  Tics2Secs := ((Tics + 9) * SecsFreq) div TicsFreq;
end;

Function Secs2Tics(Secs : LongInt) : LongInt;
  { Returns tics value for Secs seconds }
begin
  Secs2Tics := (Secs * TicsFreq) div SecsFreq;
end;

Procedure NewTimer(var ET : EventTimer; Tics : LongInt);
  { Returns a set EventTimer that will expire in Tics }
begin
  { Max acceptable value is 24 hours }
  if Tics > TicsPerDay then
    Tics := TicsPerDay;

  with ET do begin
    StartTics := BiosTics^;
    ExpireTics := StartTics + Tics;
  end;
end;

Procedure NewTimerSecs(var ET : EventTimer; Secs : LongInt);
  { Returns a set EventTimer }
begin
  NewTimer(ET, Secs2Tics(Secs));
end;

Function TimerExpired(ET : EventTimer) : Boolean;
  {-Returns True if ET has expired}
var
  CurTics : LongInt;
begin
  with ET do begin
    {Get current tics; assume timer has expired}
    CurTics := BiosTics^;
    TimerExpired := True;

    {Check normal expiration}
    if CurTics > ExpireTics then
      Exit;
    {Check wrapped CurTics}
    if (CurTics < StartTics) and ((CurTics + TicsPerDay) > ExpireTics) then
      Exit;

    {If we get here, timer hasn't expired yet}
    TimerExpired := False;
  end;
end;

Function ElapsedTime(ET : EventTimer) : LongInt;
  {-Returns elapsed time, in tics, for this timer}
var
  CurTics : LongInt;
begin
  with ET do begin
    CurTics := BiosTics^;
    if CurTics >= StartTics then
      {No midnight wrap yet}
      ElapsedTime := CurTics - StartTics
    else
      {Got a midnight wrap, account for it}
      ElapsedTime := (TicsPerDay - StartTics) + CurTics;
  end;
end;

Function ElapsedTimeInSecs(ET : EventTimer) : LongInt;
  {-Returns elapsed time, in seconds, for this timer}
begin
  ElapsedTimeInSecs := Tics2Secs(ElapsedTime(ET));
end;

Function ElapsedTimeInMSecs(ET : EventTimer) : LongInt;
  {-Returns elapsed time, in milliseconds, for this timer}
begin
  ElapsedTimeInMSecs := ElapsedTime(ET) * 55;
end;

Function RemainingTime(ET : EventTimer) : LongInt;
  {-Returns remaining time, in tics, for this timer}
var
  CurTics : LongInt;
  RemainingTics : LongInt;
begin
  with ET do begin
    CurTics := BiosTics^;
    if CurTics >= StartTics then
      {No midnight wrap yet}
      RemainingTics := ExpireTics - CurTics
    else
      {Got a midnight wrap, account for it}
      RemainingTics := (ExpireTics - TicsPerDay) - CurTics;
  end;
  if RemainingTics < 0 then
    RemainingTime := 0
  else
    RemainingTime := RemainingTics;
end;

Function RemainingTimeInSecs(ET : EventTimer) : LongInt;
  {-Returns remaining time, in seconds, for this timer}
begin
  RemainingTimeInSecs := Tics2Secs(RemainingTime(ET));
end;

Function RemainingTimeInMSecs(ET : EventTimer) : LongInt;
  {-Returns remaining time (in milliseconds) for this timer}
begin
  RemainingTimeInMSecs := RemainingTime(ET) * 55;
end;

Procedure DelayTics(Tics : LongInt);
  {-Delay for Tics tics}
var
  ET : EventTimer;
begin
  if Tics <= 0 then
    Exit
  else if Tics > TicsPerDay then
    Tics := TicsPerDay;

  NewTimer(ET, Tics);
  repeat
  until TimerExpired(ET);
end;

Procedure Delay(Ms : Word);
  {-Delay for specified number of milliseconds}
var
  DummyLow : ^Word;  {absolute $0 : $0;}
  InitTicks : Word;
  Counts : Word;
  Done : Boolean;
  T : Word;
  Extra : Word;
begin
  DummyLow := Ptr($40, 0);
  for T := 1 to Ms do begin
    InitTicks := DummyLow^;
    Counts := 0;
    repeat
      Extra := 0;
      repeat
        Inc(Extra);
      until (Extra = ExtraLoop) or (DummyLow^ <> InitTicks);
      Inc(Counts);
      Done := (Counts = CountsPerMs) or (DummyLow^ <> InitTicks);
    until Done;
  end;
end;

Procedure CalibrateDelay;
  {-Delay calibration routine, similar to CRT unit}
var
  InitTics : Word;
  Counts : Word;
  Done : Boolean;
  Extra : Word;
begin
  inline($FB); {sti}
  CountsPerMs := 65535; {Upper limit for counts per tic}

  {Wait for tick to change}
  InitTics := BiosTicsLow^;
  repeat
  until BiosTicsLow^ <> InitTics;

  {Now count until it changes again or it reaches a limit}
  InitTics := BiosTicsLow^;
  Counts := 0;
  repeat
    Extra := 0;
    repeat
      Inc(Extra);
    until (Extra = ExtraLoop) or (BiosTicsLow^ <> InitTics);
    Inc(Counts);
    Done := (Counts = CountsPerMs) or (BiosTicsLow^ <> InitTics);
  until Done;

  {Convert to counts per millisecond}
  CountsPerMs := Counts div 55;
end;

begin {TimerUnit}
  BiosTics := Ptr($40, $6C);
  BiosTicsLow := Ptr($40, $6C);
  CalibrateDelay;  {initializes CountsPerMs for Delay}
end. {TimerUnit}